import pandas as pd
import talib.abstract as ta


def atr(
    dataframe: pd.DataFrame,
    window: int = 20,
    prefix: str = "",
    inplace: bool = False,
) -> dict[str, pd.Series] | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        window (int, optional): _description_. Defaults to 20.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.

    Raises:
        RuntimeError: _description_

    Returns:
        dict[str, pd.Series] | pd.DataFrame: _description_
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )

    i = ta.ATR(dataframe, timeperiod=window)

    if inplace:
        dataframe[f"{prefix}atr"] = i
        return dataframe

    return i



from .candlestick import cdl_direction
from .talib import cdl_3blackcrows, cdl_3whitesoldiers, cdl_pattern


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    obj.cdl_direction = cdl_direction

    obj.cdl_3blackcrows = cdl_3blackcrows
    obj.cdl_3whitesoldiers = cdl_3whitesoldiers
    obj.cdl_pattern = cdl_pattern



def plot_merge(source: dict, *updates: list[dict]) -> dict:
    """Merge multiple update plot config to source config

    Args:
        source (dict): _description_
        *updates (list[dict]): _description_

    Returns:
        dict: Merged config
    """
    for update in updates:
        for k, v in update.items():
            if k not in source:
                source[k] = v
                continue
            if isinstance(source[k], list):
                source[k].extend(v)
                continue
            if isinstance(source[k], list):
                s = source[k]
                for k1, v1 in v:
                    s[k1] = v1
                continue
    return source



import pandas as pd


def series_init(
    series: pd.Series | str = "close",
    dataframe: pd.DataFrame = None,
    inplace: bool = False,
):

    if dataframe is None:
        if __debug__:
            if not isinstance(series, pd.Series):
                raise RuntimeError(
                    f"Series type '{type(series)}' is not instance of pandas.Series"
                )

            if inplace:
                raise RuntimeError("DataFrame isnot set when inplace=True")
    else:
        if __debug__:
            if not isinstance(dataframe, pd.DataFrame):
                raise RuntimeError(
                    f"DataFrame type '{type(dataframe)}' "
                    "is not instance of pandas.DataFrame"
                )

            if not isinstance(series, str):
                raise RuntimeError(
                    f"Series type {type(series)} is not string of column name"
                )

        series = dataframe[series]
    return series


def series_indicator_inject(fn):
    def __call(data, *args, **kwargs):
        if isinstance(data, pd.DataFrame):
            return fn(*args, **kwargs, dataframe=data)

        if isinstance(data, pd.Series):
            return fn(data, *args, **kwargs)

        raise RuntimeError(f"Indicator parameter type {type(data)} is invalid")

    return __call


def pandas_inject(obj: object | None = None):
    if obj is None:
        from pandas.core.base import PandasObject

        obj = PandasObject

    obj.diff = series_indicator_inject(diff)
    obj.above = series_indicator_inject(above)
    obj.below = series_indicator_inject(below)
    obj.crossover = series_indicator_inject(crossover)
    obj.crossunder = series_indicator_inject(crossunder)


def diff(series1: pd.Series, series2: pd.Series, **kwargs) -> pd.Series:
    """Difference between 2 series

    Args:
        series1 (pd.Series): _description_
        series2 (pd.Series): _description_

    Returns:
        pd.Series: Diff of 2 series
    """
    return series1 - series2


def above(series1: pd.Series, series2: pd.Series, **kwargs) -> pd.Series:
    """Check a Series is above another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: True series1 is above series2 else False
    """
    diffed = diff(series1, series2)
    return diffed.apply(lambda v: True if v > 0 else False).astype(bool)


def below(series1: pd.Series, series2: pd.Series, **kwargs) -> pd.Series:
    """Check a Series is below another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: True series1 is below series2 else False
    """
    diffed = diff(series1, series2)
    return diffed.apply(lambda v: True if v < 0 else False).astype(bool)


def crossover(series1: pd.Series, series2: pd.Series, **kwargs) -> pd.Series:
    """Check if a Series cross over another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: True if series1 cross over series2 else False
    """
    below1 = below(series1, series2).shift(1)
    above0 = above(series1, series2)
    return below1 & above0


def crossunder(series1: pd.Series, series2: pd.Series, **kwargs) -> pd.Series:
    """Check if a Series cross under another Series

    Args:
        series1 (pd.Series): first Series
        series2 (pd.Series): second Series

    Returns:
        pd.Series: True if series1 cross under series2 else False
    """
    above1 = above(series1, series2).shift(1)
    below0 = below(series1, series2)
    return below0 & above1



import pandas as pd
import talib.abstract as ta

from ..series import series_init


def rsi(
    series: pd.Series | str = "close",
    window: int = None,
    dataframe: pd.DataFrame = None,
    prefix: str = "",
    inplace: bool = False,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        series (pd.Series | str, optional): _description_. Defaults to "close".
        window (int, optional): _description_. Defaults to None.
        dataframe (pd.DataFrame, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "".
        inplace (bool, optional): _description_. Defaults to False.

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    # Validation & init
    if __debug__:
        if window is None or window <= 0:
            raise RuntimeError(f"Window {window} is invalid")
    series = series_init(series=series, dataframe=dataframe, inplace=inplace)

    # Indicator
    i = ta.RSI(series, timeperiod=window, **kwargs)

    if inplace:
        dataframe[f"{prefix}rsi"] = i
        return dataframe

    return i



from typing import Literal

import pandas as pd
import talib.abstract as ta

from ..series import series_init
from ..utils import talib_ma_mode


def bollinger_bands(
    series: pd.Series | str = "close",
    window: int = None,
    std: int | float = None,
    dataframe: pd.DataFrame = None,
    ma_mode: Literal["ema", "sma"] = "sma",
    prefix: str = "bb_",
    inplace: bool = False,
    **kwargs,
) -> dict[str, pd.Series] | pd.DataFrame:

    if __debug__:
        if window is None or window <= 0:
            raise RuntimeError(f"Window {window} is invalid")
    series = series_init(series=series, dataframe=dataframe, inplace=inplace)

    i_upper, i_basis, i_lower = ta.BBANDS(
        series,
        timeperiod=window,
        nbdevup=std,
        nbdevdn=std,
        matype=talib_ma_mode(ma_mode),
    )

    # Result is inplace or new dict
    result = dataframe if inplace else {}
    result[f"{prefix}upper"] = i_upper
    result[f"{prefix}basis"] = i_basis
    result[f"{prefix}lower"] = i_lower
    return result



import pandas as pd

from .helper import plot_merge


def plot_ichimoku(
    dataframe: pd.DataFrame,
    tenkan_sen="tenkan_sen",
    kijun_sen="kijun_sen",
    senkou_span_a="senkou_span_a",
    senkou_span_b="senkou_span_b",
    chikou_span="chikou_span",
    width=1,
    tenkan_sen_color="#33BDFF",
    kijun_sen_color="#D105F5",
    senkou_span_a_color="#228B22",
    senkou_span_b_color="#FF3342",
    chikou_span_color="#F1F316",
) -> dict:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        tenkan_sen (str, optional): _description_. Defaults to "tenkan_sen".
        kijun_sen (str, optional): _description_. Defaults to "kijun_sen".
        senkou_span_a (str, optional): _description_. Defaults to "senkou_span_a".
        senkou_span_b (str, optional): _description_. Defaults to "senkou_span_b".
        chikou_span (str, optional): _description_. Defaults to "chikou_span".
        width (int, optional): _description_. Defaults to 1.
        tenkan_sen_color (str, optional): _description_. Defaults to "#33BDFF".
        kijun_sen_color (str, optional): _description_. Defaults to "#D105F5".
        senkou_span_a_color (str, optional): _description_. Defaults to "#228B22".
        senkou_span_b_color (str, optional): _description_. Defaults to "#FF3342".
        chikou_span_color (str, optional): _description_. Defaults to "#F1F316".

    Returns:
        dict: _description_
    """
    return dict(
        items=[
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[tenkan_sen],
                name=tenkan_sen,
                mode="lines",
                line=dict(color=tenkan_sen_color, width=width),
            ),
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[kijun_sen],
                name=kijun_sen,
                mode="lines",
                line=dict(color=kijun_sen_color, width=width),
            ),
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[senkou_span_a],
                name=senkou_span_a,
                mode="lines",
                line=dict(color=senkou_span_a_color, width=width),
            ),
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[senkou_span_b],
                name=senkou_span_b,
                mode="lines",
                fill="tonexty",
                line=dict(color=senkou_span_b_color, width=width),
            ),
            dict(
                type="scatter",
                x=dataframe.index,
                y=dataframe[chikou_span],
                name=chikou_span,
                mode="lines",
                line=dict(color=chikou_span_color, width=width),
            ),
        ]
    )


def plot_line(
    series: pd.Series,
    color: str = "#ffee58",
    width: int = 1,
    name=None,
    mode="lines",
    **kwargs,
) -> dict:
    """_summary_

    Args:
        series (pd.Series): _description_
        color (str, optional): _description_. Defaults to "#ffee58".
        width (int, optional): _description_. Defaults to 1.
        name (_type_, optional): _description_. Defaults to None.
        mode (str, optional): _description_. Defaults to "lines".

    Returns:
        dict: _description_
    """
    return dict(
        items=[
            dict(
                type="scatter",
                x=series.index,
                y=series,
                line=dict(color=color, width=width),
                name=name or series.name,
                mode=mode,
                **kwargs,
            )
        ]
    )


def plot_lines(
    *serieses: list[pd.Series],
    color: str = "#ffee58",
    width: int = 1,
    name=None,
    mode="lines",
    **kwargs,
) -> dict:
    """_summary_

    Args:
        color (str, optional): _description_. Defaults to "#ffee58".
        width (int, optional): _description_. Defaults to 1.
        name (_type_, optional): _description_. Defaults to None.
        mode (str, optional): _description_. Defaults to "lines".

    Returns:
        dict: _description_
    """
    result = {}
    for series in serieses:
        plot_merge(
            result,
            plot_line(
                series=series,
                color=color,
                width=width,
                name=name,
                mode=mode,
                **kwargs,
            ),
        )
    return result


def plot_mark(
    series: pd.Series,
    color: str = "#ffee58",
    width: int = 1,
    mode="markers",
    name=None,
    **kwargs,
) -> dict:
    """_summary_

    Args:
        series (pd.Series): _description_
        color (str, optional): _description_. Defaults to "#ffee58".
        width (int, optional): _description_. Defaults to 1.
        mode (str, optional): _description_. Defaults to "markers".
        name (_type_, optional): _description_. Defaults to None.

    Returns:
        dict: _description_
    """
    return plot_line(
        series=series,
        color=color,
        width=width,
        mode=mode,
        name=name,
        **kwargs,
    )


def plot_candlestick(
    dataframe: pd.DataFrame,
    name: str = "Candlestick",
    width: int = 1,
    increasing_line_color="#26c6da",
    decreasing_line_color="#ab47bc",
    row: int = 1,
    col: int = 1,
    **kwargs,
) -> dict:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str, optional): _description_. Defaults to "Candle highlight".
        width (int, optional): _description_. Defaults to 1.
        increasing_line_color (str, optional): _description_. Defaults to "#26c6da".
        decreasing_line_color (str, optional): _description_. Defaults to "#ab47bc".

    Returns:
        dict: _description_
    """
    return dict(
        items=[
            dict(
                type="candlestick",
                x=dataframe.index,
                open=dataframe["open"],
                high=dataframe["high"],
                low=dataframe["low"],
                close=dataframe["close"],
                name=name,
                line=dict(width=width),
                increasing_line_color=increasing_line_color,
                decreasing_line_color=decreasing_line_color,
                hoverinfo="text",
                hovertext=name,
                row=row,
                col=col,
                **kwargs,
            ),
        ]
    )



import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from lettrade.data import DataFeed
from lettrade.plot import BotPlotter, PlotColor


class PlotlyBotPlotter(BotPlotter):
    """Class help to plot `lettrade`"""

    figure: go.Figure = None

    _data_shape: dict

    def stop(self):
        """stop plotter"""

    def default_config(self):
        # --- Data
        groups = []
        for data in self.datas:
            groups.append(
                dict(
                    id=data.name,
                    type="data",
                    data=data,
                    height=1,
                    items=[
                        dict(
                            type="candlestick",
                            show_orders=data is self.data,
                            show_positions=data is self.data,
                            update_xaxes=dict(
                                title_text=data.name,
                                rangeslider_visible=False,
                                # showspikes=True,
                                # spikemode="across",
                                mirror=True,
                                ticks="outside",
                                showline=True,
                                # linecolor="",
                            ),
                            update_yaxes=dict(
                                title_text="Price $",
                                # autorange=True,
                                # fixedrange=False,
                                # showspikes=True,
                                # spikemode="across",
                                mirror=True,
                                ticks="outside",
                                showline=True,
                            ),
                        )
                    ],
                )
            )

        # --- Equity
        groups.append(
            dict(
                id="equity",
                type="equity",
                height=0.5,
            )
        )

        # --- Buttons
        buttons = [dict(step="all")]
        match self.data.timeframe.unit:
            case "m":
                count = 1
                step = "day"
            case "h":
                count = 7
                step = "day"
            case _:
                count = 1
                step = "month"
        buttons.extend(
            [
                dict(
                    count=6 * count,
                    label=f"{6*count} {step}",
                    step=step,
                    stepmode="backward",
                ),
                dict(
                    count=3 * count,
                    label=f"{3*count} {step}",
                    step=step,
                    stepmode="backward",
                ),
                dict(
                    count=count,
                    label=f"{count} {step}",
                    step=step,
                    stepmode="backward",
                ),
            ]
        )
        return dict(
            groups=groups,
            # --- Params
            params=dict(
                shared_xaxes=True,
                vertical_spacing=0.03,
            ),
            # --- Layout
            layout=dict(
                xaxis=dict(
                    rangeselector=dict(
                        bgcolor="#282a36",
                        activecolor="#5b5b66",
                        buttons=buttons,
                    ),
                ),
                yaxis=dict(
                    autorange=True,
                    fixedrange=False,
                ),
                title=dict(
                    text=str(self.strategy),
                    font=dict(size=24),
                    x=0.5,
                    xref="paper",
                ),
                # autosize=False,
                # width=800,
                # height=1_000 * plot_rows,
                modebar_add=[
                    "v1hovermode",
                    "hoverclosest",
                    "hovercompare",
                    "togglehover",
                    "togglespikelines",
                    "drawline",
                    "drawopenpath",
                    "drawclosedpath",
                    "drawcircle",
                    "drawrect",
                    "eraseshape",
                ],
                # template="plotly_dark",
                hovermode="x unified",
            ),
        )

    def _config_standard(self, config: dict):
        shapes = dict()
        for i, group in enumerate(config["groups"]):
            if group["type"] == "data":
                rows = max(item.get("row", 1) for item in group["items"])
                cols = max(item.get("col", 1) for item in group["items"])
            else:
                rows = group.get("row", 1)
                cols = group.get("col", 1)

            shapes[group["id"]] = dict(group_index=i, rows=rows, cols=cols)

        rows_total = sum(shape["rows"] for shape in shapes.values())
        cols_total = sum(shape["cols"] for shape in shapes.values())

        params: dict = config.setdefault("params", dict())
        params.update(
            rows=rows_total,
            row_heights=[1] * (rows_total - 1) + [0.5],
        )

        config.update(
            params=params,
            shapes=shapes,
            rows_total=rows_total,
            cols_total=cols_total,
        )
        return config

    def _strategy_config(self, config: dict):
        strategy_config: dict = self.strategy.plot(*self.datas)

        # Plot scatter/trace
        if "items" in strategy_config:
            shape = config["shapes"][self.data.name]
            data_group = config["groups"][shape["group_index"]]
            data_items: list = data_group.setdefault("items", [])
            data_items.extend(strategy_config["items"])

        for data in self.datas:
            dname = data.name
            if dname not in strategy_config:
                continue

            data_shape = config["shapes"][data.name]
            data_group = config["groups"][data_shape["group_index"]]
            data_config = strategy_config[dname]

            data_items: list = data_group.setdefault("items", [])
            data_items.extend(data_config["items"])

        return config

    def _config_render(self, config: dict):
        rows = 0
        for group in config["groups"]:
            group_shape = config["shapes"][group["id"]]

            if group["type"] == "data":
                for item in group["items"]:
                    row = rows + item.pop("row", 1)
                    col = item.pop("col", 1)
                    match item["type"]:
                        case "candlestick":
                            self._plot_candlestick(
                                data=group["data"],
                                row=row,
                                col=col,
                                **item,
                            )
                        case "scatter":
                            self._plot_scatter(
                                data=group["data"],
                                row=row,
                                col=col,
                                **item,
                            )
            elif group["type"] == "equity":
                row = rows + group.pop("row", 1)
                col = group.pop("col", 1)

                self._plot_equity(row=row, col=col, **group)

            rows += group_shape["rows"]

    def _plot_candlestick(
        self,
        data: DataFeed,
        x=None,
        open=None,
        high=None,
        low=None,
        close=None,
        update_xaxes: dict | None = None,
        update_yaxes: dict | None = None,
        show_orders: bool = False,
        show_positions: bool = False,
        row: int = 1,
        col: int = 1,
        type: str | None = None,
        **kwargs,
    ):
        self.figure.add_trace(
            go.Candlestick(
                x=x or data.index,
                open=open or data["open"],
                high=high or data["high"],
                low=low or data["low"],
                close=close or data["close"],
                **kwargs,
            ),
            row=row,
            col=col,
        )
        if update_xaxes is not None:
            self.figure.update_xaxes(row=row, col=col, **update_xaxes)
        if update_yaxes is not None:
            self.figure.update_yaxes(row=row, col=col, **update_yaxes)
        if show_orders:
            self._plot_orders(data=data, row=row, col=col)
        if show_positions:
            self._plot_positions(data=data, row=row, col=col)

    def _plot_scatter(
        self,
        data: DataFeed,
        x: pd.Series,
        y: pd.Series,
        row: int = 1,
        col: int = 1,
        type: str | None = None,
        **kwargs,
    ):
        self.figure.add_scatter(x=x, y=y, row=row, col=col, **kwargs)

    def _plot_equity(
        self,
        row: int = 1,
        col: int = 1,
        type: str | None = None,
        **kwargs,
    ):
        equities = self.account._equities

        # Filter equities in data range only
        start_dt = self.data.index[0]
        stop_dt = self.data.index[-1]
        equities = {k: e for k, e in equities.items() if k > start_dt and k < stop_dt}

        # Axis
        x = list(equities.keys())
        y = list(equities.values())

        self.figure.add_trace(
            go.Scatter(
                x=x,
                y=y,
                line=dict(color="#ff9900", width=2),
                # showlegend=False,
                name="Equity",
                # fill="toself",
            ),
            row=row,
            col=col,
        )
        self.figure.update_yaxes(
            title_text="Equity",
            row=row,
            col=col,
        )

    def _plot_orders(
        self,
        data: DataFeed,
        row: int = 1,
        col: int = 1,
    ):
        orders = list(self.exchange.history_orders.values()) + list(
            self.exchange.orders.values()
        )

        # Filter order in data range only
        start_dt = data.index[0]
        stop_dt = data.index[-1]
        orders = [o for o in orders if o.placed_at > start_dt and o.placed_at < stop_dt]

        # TODO: test using order/sl/tp line for performance
        for order in orders:
            x = [order.placed_at]
            y = [order.place_price]

            hovertemplate = (
                "<br>"
                # f"Order id: {order.id}<br>"
                "Index: %{x}<br>"
                f"At: {order.placed_at}<br>"
                "Price: %{y}<br>"
                f"Size: {order.size}<br>"
            )
            if order.sl:
                hovertemplate += f"SL: {order.sl}<br>"
            if order.tp:
                hovertemplate += f"TP: {order.tp}<br>"

            self.figure.add_scatter(
                x=x,
                y=y,
                hovertemplate=hovertemplate,
                mode="markers",
                name=f"Order-{order.id}",
                marker=dict(
                    symbol="line-ew-open",
                    size=10,
                    color=(PlotColor.GREEN if order.is_long else PlotColor.RED),
                ),
                showlegend=False,
                row=row,
                col=col,
            )

    def _plot_positions(
        self,
        data: DataFeed,
        row: int = 1,
        col: int = 1,
    ):
        positions = list(self.exchange.history_positions.values()) + list(
            self.exchange.positions.values()
        )

        # Filter equities in data range only
        start_dt = data.index[0]
        stop_dt = data.index[-1]
        _positions = []
        for p in positions:
            if p.entry_at < start_dt or p.entry_at > stop_dt:
                continue
            if p.exit_at:
                if p.exit_at < start_dt or p.exit_at > stop_dt:
                    continue
            _positions.append(p)
        positions = _positions

        for position in positions:
            x = [position.entry_at]
            y = [position.entry_price]
            customdata = [["Entry", position.size, position.pl, position.fee]]
            if position.exit_at:
                x.append(position.exit_at)
                y.append(position.exit_price)
                customdata.append(["Exit", position.size, position.pl, position.fee])

            color = PlotColor.CYAN if position.is_long else PlotColor.LIGHT_PINK
            self.figure.add_scatter(
                x=x,
                y=y,
                customdata=customdata,
                hovertemplate=(
                    "<br>"
                    "At: %{x}<br>"
                    "Price: %{y}<br>"
                    "Status: %{customdata[0]}<br>"
                    "Size: %{customdata[1]}<br>"
                    "PL: %{customdata[2]:.2f}$<br>"
                    "Fee: %{customdata[3]:.2f}$<br>"
                ),
                mode="lines+markers",
                name=f"Position-{position.id}",
                marker=dict(
                    symbol="circle-dot",
                    size=10,
                    color=color,
                ),
                line=dict(
                    color=color,
                    width=1,
                    dash="dash",
                ),
                showlegend=False,
                row=row,
                col=col,
            )

    def load(self):
        """Load plot config from `Strategy.plot()` and setup candlestick/equity"""

        config = self.default_config()
        config = self._config_standard(config)
        config = self._strategy_config(config)

        # Init
        params: dict = config.setdefault("params", dict())
        self.figure = make_subplots(**params)

        self._config_render(config=config)

        self.figure.update_layout(**config["layout"])

    def plot(self, jump: dict | None = None, **kwargs):
        """Plot `equity`, `orders`, and `positions` then show"""
        if jump is not None:
            self.jump(**jump)
        elif self.figure is None:
            self.load()
        else:
            if self.jump_reset():
                self.load()

        params = dict(layout_xaxis_rangeslider_visible=False)
        params.update(**kwargs)
        self.figure.update(**params)

        self.figure.show()



import pandas as pd
import talib.abstract as ta


def cdl_3blackcrows(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="3blackcrows",
        name=name,
        prefix=prefix,
        inplace=inplace,
        **kwargs,
    )


def cdl_3whitesoldiers(
    dataframe: pd.DataFrame,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): _description_
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    return cdl_pattern(
        dataframe=dataframe,
        pattern="3whitesoldiers",
        name=name,
        prefix=prefix,
        inplace=inplace,
        **kwargs,
    )


def cdl_pattern(
    dataframe: pd.DataFrame,
    pattern: str,
    name: str | None = None,
    prefix: str = "cdl_",
    inplace: bool = False,
    **kwargs,
) -> pd.Series | pd.DataFrame:
    """_summary_

    Args:
        dataframe (pd.DataFrame): pandas.DataFrame with ohlcv
        pattern (str): TA-Lib candle pattern name. Ex: `3whitesoldiers`, `3blackcrows`
        name (str | None, optional): _description_. Defaults to None.
        prefix (str, optional): _description_. Defaults to "cdl_".
        inplace (bool, optional): _description_. Defaults to False.

    Raises:
        RuntimeError: _description_

    Returns:
        pd.Series | pd.DataFrame: _description_
    """
    if __debug__:
        if not isinstance(dataframe, pd.DataFrame):
            raise RuntimeError(
                f"dataframe type '{type(dataframe)}' "
                "is not instance of pandas.DataFrame"
            )

    i = getattr(ta, f"CDL{pattern.upper()}")(dataframe, **kwargs)

    if inplace:
        dataframe[name or f"{prefix}{pattern.lower()}"] = i
        return dataframe

    return i
